pub mod prime_utils {
    use num::{BigUint, One};
    use once_cell::sync::Lazy;

    #[must_use]
    pub fn log_2(x: &BigUint) -> u64 {
        x.bits() - 1
    }

    #[must_use]
    pub fn is_prime(number: &BigUint) -> bool {
        const ZERO: BigUint = BigUint::ZERO;
        static TWO: Lazy<BigUint> = Lazy::new(|| BigUint::from(2u8));

        if &BigUint::one() >= number {
            return false;
        }
        if &BigUint::from(4u8) > number {
            return true;
        }
        if number % &*TWO == ZERO {
            return false;
        }

        // number = 2^a - 1
        // a = log2(number + 1)
        let a = log_2(&(number + 1u8));

        //if number isn't a mersenne number
        if TWO.pow(a as u32) - BigUint::one() != *number {
            let mut i = BigUint::one();

            let sqrtnum = number.sqrt() + BigUint::one(); //fake ceil function

            loop {
                i += &*TWO; //we begin checking 3 and then continue with every second number as every even number is divisible by 2 and we already checked that
                if number % &i == ZERO {
                    return false;
                }
                if i >= sqrtnum {
                    return true;
                }
            }
        }

        //check if it's a mersenne prime
        let mut last = BigUint::from(4u8);

        for _i in 2..a {
            last = (last.pow(2) - &*TWO) % number;
        }

        last == BigUint::ZERO
    }
}
